@@ -51,8 +51,11 @@ from dataclasses import dataclass
 from .compat import SerdeError, SerdeSkip
 from .core import (
     AdjacentTagging,
+    Coerce,
     ExternalTagging,
     InternalTagging,
+    NoCheck,
+    Strict,
     Untagged,
     field,
     init,
@@ -69,6 +72,9 @@ __all__ = (
     "ExternalTagging",
     "InternalTagging",
     "Untagged",
+    "NoCheck",
+    "Strict",
+    "Coerce",
     "field",
     "default_deserializer",
     "deserialize",
