@@ -13,26 +13,32 @@ from typing import Any, Callable, Dict, List, Optional, Type, TypeVar
 import jinja2
 
 from .compat import (
+    Literal,
     SerdeError,
     SerdeSkip,
+    UserError,
     find_generic_arg,
     get_generic_arg,
     get_origin,
     has_default,
     has_default_factory,
+    is_any,
     is_bare_dict,
     is_bare_list,
     is_bare_set,
     is_bare_tuple,
+    is_datetime,
     is_dict,
     is_enum,
     is_generic,
     is_list,
     is_literal,
+    is_new_type_primitive,
     is_none,
     is_opt,
     is_primitive,
     is_set,
+    is_str_serializable,
     is_tuple,
     is_union,
     iter_literals,
@@ -45,19 +51,23 @@ from .core import (
     FROM_DICT,
     FROM_ITER,
     SERDE_SCOPE,
+    TYPE_CHECK,
     UNION_DE_PREFIX,
-    DateTimeTypes,
     DefaultTagging,
     Field,
+    NoCheck,
     SerdeScope,
-    StrSerializableTypes,
     Tagging,
+    TypeCheck,
     add_func,
+    coerce,
     ensure,
     fields,
+    is_instance,
     literal_func_name,
     logger,
     raise_unsupported_type,
+    render_type_check,
     union_func_name,
 )
 from .numpy import (
@@ -128,6 +138,7 @@ def deserialize(
     reuse_instances_default: bool = True,
     deserializer: Optional[DeserializeFunc] = None,
     tagging: Tagging = DefaultTagging,
+    type_check: TypeCheck = NoCheck,
     **kwargs,
 ):
     """
@@ -199,7 +210,7 @@ def deserialize(
         # Create a scope storage used by serde.
         # Each class should get own scope. Child classes can not share scope with parent class.
         # That's why we need the "scope.cls is not cls" check.
-        scope: SerdeScope = getattr(cls, SERDE_SCOPE, None)
+        scope: Optional[SerdeScope] = getattr(cls, SERDE_SCOPE, None)
         if scope is None or scope.cls is not cls:
             scope = SerdeScope(cls, reuse_instances_default=reuse_instances_default)
             setattr(cls, SERDE_SCOPE, scope)
@@ -208,12 +219,18 @@ def deserialize(
         g['cls'] = cls
         g['serde_scope'] = scope
         g['SerdeError'] = SerdeError
+        g['UserError'] = UserError
         g['raise_unsupported_type'] = raise_unsupported_type
         g['typename'] = typename  # used in union functions
         g['ensure'] = ensure
         g['typing'] = typing
+        g['Literal'] = Literal
         g['from_obj'] = from_obj
         g['get_generic_arg'] = get_generic_arg
+        g['is_instance'] = is_instance
+        g['TypeCheck'] = TypeCheck
+        g['NoCheck'] = NoCheck
+        g['coerce'] = coerce
         if deserialize:
             g['serde_custom_class_deserializer'] = functools.partial(
                 serde_custom_class_deserializer, custom=deserializer
@@ -221,7 +238,7 @@ def deserialize(
 
         # Collect types used in the generated code.
         for typ in iter_types(cls):
-            if typ is cls or (is_primitive(typ) and not is_enum(typ)):
+            if typ is cls or (is_primitive(typ) and not is_enum(typ) and not is_new_type_primitive(typ)):
                 continue
             if is_generic(typ):
                 g[typename(typ)] = get_origin(typ)
@@ -250,8 +267,9 @@ def deserialize(
             if f.deserializer:
                 g[f.deserializer.name] = f.deserializer
 
-        add_func(scope, FROM_ITER, render_from_iter(cls, deserializer), g)
-        add_func(scope, FROM_DICT, render_from_dict(cls, rename_all, deserializer), g)
+        add_func(scope, FROM_ITER, render_from_iter(cls, deserializer, type_check), g)
+        add_func(scope, FROM_DICT, render_from_dict(cls, rename_all, deserializer, type_check), g)
+        add_func(scope, TYPE_CHECK, render_type_check(cls), g)
 
         logger.debug(f'{typename(cls)}: {SERDE_SCOPE} {scope}')
 
@@ -316,14 +334,9 @@ def from_obj(c: Type, o: Any, named: bool, reuse_instances: bool):
             return None
         if is_deserializable(c):
             serde_scope: SerdeScope = getattr(c, SERDE_SCOPE)
-            if named:
-                return serde_scope.funcs[FROM_DICT](
-                    cls=c, maybe_generic=maybe_generic, data=o, reuse_instances=reuse_instances
-                )
-            else:
-                return serde_scope.funcs[FROM_ITER](
-                    cls=c, maybe_generic=maybe_generic, data=o, reuse_instances=reuse_instances
-                )
+            func_name = FROM_DICT if named else FROM_ITER
+            res = serde_scope.funcs[func_name](c, maybe_generic=maybe_generic, data=o, reuse_instances=reuse_instances)
+            return res
         elif is_opt(c):
             if o is None:
                 return None
@@ -342,31 +355,38 @@ def from_obj(c: Type, o: Any, named: bool, reuse_instances: bool):
             if is_bare_list(c):
                 return [e for e in o]
             else:
-                return [thisfunc(type_args(c)[0], e) for e in o]
+                res = [thisfunc(type_args(c)[0], e) for e in o]
+                return res
         elif is_set(c):
             if is_bare_set(c):
                 return set(e for e in o)
             else:
-                return set(thisfunc(type_args(c)[0], e) for e in o)
+                res = set(thisfunc(type_args(c)[0], e) for e in o)
+                return res
         elif is_tuple(c):
             if is_bare_tuple(c):
                 return tuple(e for e in o)
             else:
-                return tuple(thisfunc(type_args(c)[i], e) for i, e in enumerate(o))
+                res = tuple(thisfunc(type_args(c)[i], e) for i, e in enumerate(o))
+                return res
         elif is_dict(c):
             if is_bare_dict(c):
                 return {k: v for k, v in o.items()}
             else:
-                return {thisfunc(type_args(c)[0], k): thisfunc(type_args(c)[1], v) for k, v in o.items()}
+                res = {thisfunc(type_args(c)[0], k): thisfunc(type_args(c)[1], v) for k, v in o.items()}
+                return res
         elif is_numpy_array(c):
             return deserialize_numpy_array_direct(c, o)
-        elif c in DateTimeTypes:
+        elif is_datetime(c):
             return c.fromisoformat(o)
-        elif c is Any:
+        elif is_any(c):
             return o
 
         return c(o)
 
+    except UserError as e:
+        raise e.inner
+
     except Exception as e:
         raise SerdeError(e)
 
@@ -507,6 +527,7 @@ class Renderer:
     cls: Optional[Type] = None
     custom: Optional[DeserializeFunc] = None  # Custom class level deserializer.
     import_numpy: bool = False
+    suppress_coerce: bool = False
 
     def render(self, arg: DeField) -> str:
         """
@@ -538,9 +559,9 @@ class Renderer:
             res = self.primitive(arg)
         elif is_union(arg.type):
             res = self.union_func(arg)
-        elif arg.type in StrSerializableTypes:
+        elif is_str_serializable(arg.type):
             res = f"({self.c_tor_with_check(arg)}) if reuse_instances else {self.c_tor(arg)}"
-        elif arg.type in DateTimeTypes:
+        elif is_datetime(arg.type):
             from_iso = f"{typename(arg.type)}.fromisoformat({arg.data})"
             res = f"({arg.data} if isinstance({arg.data}, {typename(arg.type)}) else {from_iso}) \
                     if reuse_instances else {from_iso}"
@@ -550,7 +571,10 @@ class Renderer:
             res = arg.data
         elif isinstance(arg.type, TypeVar):
             index = find_generic_arg(self.cls, arg.type)
-            res = f"from_obj(get_generic_arg(maybe_generic, {index}), {arg.data}, named={not arg.iterbased}, reuse_instances=reuse_instances)"
+            res = (
+                f"from_obj(get_generic_arg(maybe_generic, {index}), "
+                f" {arg.data}, named={not arg.iterbased}, reuse_instances=reuse_instances)"
+            )
         elif is_generic(arg.type):
             arg.type = get_origin(arg.type)
             res = self.render(arg)
@@ -564,7 +588,7 @@ class Renderer:
 
         if self.custom and not arg.deserializer:
             # Rerender the code for default deserializer.
-            default = Renderer(self.func, self.cls, None).render(arg)
+            default = Renderer(self.func, self.cls, None, suppress_coerce=self.suppress_coerce).render(arg)
             return self.custom_class_deserializer(arg, default)
         else:
             return res
@@ -608,13 +632,13 @@ class Renderer:
 
         >>> from typing import List
         >>> Renderer('foo').render(DeField(Optional[int], 'o', datavar='data'))
-        '(data["o"]) if data.get("o") is not None else None'
+        '(coerce(int, data["o"])) if data.get("o") is not None else None'
 
         >>> Renderer('foo').render(DeField(Optional[List[int]], 'o', datavar='data'))
-        '([v for v in data["o"]]) if data.get("o") is not None else None'
+        '([coerce(int, v) for v in data["o"]]) if data.get("o") is not None else None'
 
         >>> Renderer('foo').render(DeField(Optional[List[int]], 'o', datavar='data'))
-        '([v for v in data["o"]]) if data.get("o") is not None else None'
+        '([coerce(int, v) for v in data["o"]]) if data.get("o") is not None else None'
 
         >>> @deserialize
         ... class Foo:
@@ -639,10 +663,10 @@ if data.get("f") is not None else None'
 
         >>> from typing import List
         >>> Renderer('foo').render(DeField(List[int], 'l', datavar='data'))
-        '[v for v in data["l"]]'
+        '[coerce(int, v) for v in data["l"]]'
 
         >>> Renderer('foo').render(DeField(List[List[int]], 'l', datavar='data'))
-        '[[v for v in v] for v in data["l"]]'
+        '[[coerce(int, v) for v in v] for v in data["l"]]'
         """
         if is_bare_list(arg.type):
             return f'list({arg.data})'
@@ -655,10 +679,10 @@ if data.get("f") is not None else None'
 
         >>> from typing import Set
         >>> Renderer('foo').render(DeField(Set[int], 'l', datavar='data'))
-        'set(v for v in data["l"])'
+        'set(coerce(int, v) for v in data["l"])'
 
         >>> Renderer('foo').render(DeField(Set[Set[int]], 'l', datavar='data'))
-        'set(set(v for v in v) for v in data["l"])'
+        'set(set(coerce(int, v) for v in v) for v in data["l"])'
         """
         if is_bare_set(arg.type):
             return f'set({arg.data})'
@@ -673,13 +697,15 @@ if data.get("f") is not None else None'
         >>> @deserialize
         ... class Foo: pass
         >>> Renderer('foo').render(DeField(Tuple[str, int, List[int], Foo], 'd', datavar='data'))
-        '(data["d"][0], data["d"][1], [v for v in data["d"][2]], \
+        '(coerce(str, data["d"][0]), coerce(int, data["d"][1]), \
+[coerce(int, v) for v in data["d"][2]], \
 Foo.__serde__.funcs[\\'foo\\'](data=data["d"][3], reuse_instances=reuse_instances),)'
 
         >>> field = DeField(Tuple[str, int, List[int], Foo], 'd', datavar='data', index=0, iterbased=True)
         >>> Renderer('foo').render(field)
-        "(data[0][0], data[0][1], [v for v in data[0][2]], \
-Foo.__serde__.funcs['foo'](data=data[0][3], reuse_instances=reuse_instances),)"
+        "(coerce(str, data[0][0]), coerce(int, data[0][1]), \
+[coerce(int, v) for v in data[0][2]], Foo.__serde__.funcs['foo'](data=data[0][3], \
+reuse_instances=reuse_instances),)"
         """
         if is_bare_tuple(arg.type):
             return f'tuple({arg.data})'
@@ -696,7 +722,7 @@ Foo.__serde__.funcs['foo'](data=data[0][3], reuse_instances=reuse_instances),)"
 
         >>> from typing import List
         >>> Renderer('foo').render(DeField(Dict[str, int], 'd', datavar='data'))
-        '{k: v for k, v in data["d"].items()}'
+        '{coerce(str, k): coerce(int, v) for k, v in data["d"].items()}'
 
         >>> @deserialize
         ... class Foo: pass
@@ -714,20 +740,25 @@ Foo.__serde__.funcs['foo'](data=data[0][3], reuse_instances=reuse_instances),)"
     def enum(self, arg: DeField) -> str:
         return f'{typename(arg.type)}({self.primitive(arg)})'
 
-    def primitive(self, arg: DeField) -> str:
+    def primitive(self, arg: DeField, suppress_coerce: bool = False) -> str:
         """
         Render rvalue for primitives.
 
         >>> Renderer('foo').render(DeField(int, 'i', datavar='data'))
-        'data["i"]'
+        'coerce(int, data["i"])'
 
         >>> Renderer('foo').render(DeField(int, 'int_field', datavar='data', case='camelcase'))
-        'data["intField"]'
+        'coerce(int, data["intField"])'
 
         >>> Renderer('foo').render(DeField(int, 'i', datavar='data', index=1, iterbased=True))
-        'data[1]'
+        'coerce(int, data[1])'
         """
-        return arg.data
+        typ = typename(arg.type)
+        dat = arg.data
+        if self.suppress_coerce:
+            return dat
+        else:
+            return f'coerce({typ}, {dat})'
 
     def c_tor(self, arg: DeField) -> str:
         return f"{typename(arg.type)}({arg.data})"
@@ -768,7 +799,7 @@ def to_iter_arg(f: DeField, *args, **kwargs) -> DeField:
     return f
 
 
-def render_from_iter(cls: Type, custom: Optional[DeserializeFunc] = None) -> str:
+def render_from_iter(cls: Type, custom: Optional[DeserializeFunc] = None, type_check: TypeCheck = NoCheck) -> str:
     template = """
 def {{func}}(cls=cls, maybe_generic=None, data=None, reuse_instances = {{serde_scope.reuse_instances_default}}):
   if reuse_instances is Ellipsis:
@@ -777,14 +808,21 @@ def {{func}}(cls=cls, maybe_generic=None, data=None, reuse_instances = {{serde_s
   if data is None:
     return None
 
-  return cls(
   {% for f in fields %}
-  {{f|arg(loop.index-1)|rvalue}},
+  __{{f.name}} = {{f|arg(loop.index-1)|rvalue}}
   {% endfor %}
-  )
+
+  try:
+    return cls(
+      {% for f in fields %}
+      __{{f.name}},
+      {% endfor %}
+    )
+  except Exception as e:
+    raise UserError(e)
     """
 
-    renderer = Renderer(FROM_ITER, cls=cls, custom=custom)
+    renderer = Renderer(FROM_ITER, cls=cls, custom=custom, suppress_coerce=(not type_check.is_coerce()))
     env = jinja2.Environment(loader=jinja2.DictLoader({'iter': template}))
     env.filters.update({'rvalue': renderer.render})
     env.filters.update({'arg': to_iter_arg})
@@ -796,28 +834,48 @@ def {{func}}(cls=cls, maybe_generic=None, data=None, reuse_instances = {{serde_s
     return res
 
 
-def render_from_dict(cls: Type, rename_all: Optional[str] = None, custom: Optional[DeserializeFunc] = None) -> str:
+def render_from_dict(
+    cls: Type,
+    rename_all: Optional[str] = None,
+    custom: Optional[DeserializeFunc] = None,
+    type_check: TypeCheck = NoCheck,
+) -> str:
     template = """
-def {{func}}(cls=cls, maybe_generic=None, data=None, reuse_instances = {{serde_scope.reuse_instances_default}}):
+def {{func}}(cls=cls, maybe_generic=None, data=None,
+             reuse_instances = {{serde_scope.reuse_instances_default}}):
   if reuse_instances is Ellipsis:
     reuse_instances = {{serde_scope.reuse_instances_default}}
 
   if data is None:
     return None
 
-  rv = cls(
   {% for f in fields %}
-  {{f|arg(loop.index-1)|rvalue}},
+  __{{f.name}} = {{f|arg(loop.index-1)|rvalue}}
   {% endfor %}
-  )
+
+  try:
+    rv = cls(
+    {% for f in fields %}
+    __{{f.name}},
+    {% endfor %}
+    )
+  except Exception as e:
+    raise UserError(e)
+
+  {% if type_check.is_strict() %}
+  rv.__serde__.funcs['typecheck'](rv)
+  {% endif %}
+
   return rv
     """
 
-    renderer = Renderer(FROM_DICT, cls=cls, custom=custom)
+    renderer = Renderer(FROM_DICT, cls=cls, custom=custom, suppress_coerce=(not type_check.is_coerce()))
     env = jinja2.Environment(loader=jinja2.DictLoader({'dict': template}))
     env.filters.update({'rvalue': renderer.render})
     env.filters.update({'arg': functools.partial(to_arg, rename_all=rename_all)})
-    res = env.get_template('dict').render(func=FROM_DICT, serde_scope=getattr(cls, SERDE_SCOPE), fields=defields(cls))
+    res = env.get_template('dict').render(
+        func=FROM_DICT, serde_scope=getattr(cls, SERDE_SCOPE), fields=defields(cls), type_check=type_check
+    )
 
     if renderer.import_numpy:
         res = "import numpy\n" + res
@@ -864,7 +922,7 @@ def {{func}}(cls=cls, maybe_generic=None, data=None, reuse_instances = {{serde_s
     """
     union_name = f"Union[{', '.join([typename(a) for a in union_args])}]"
 
-    renderer = Renderer(FROM_DICT, cls=cls)
+    renderer = Renderer(FROM_DICT, cls=cls, suppress_coerce=True)
     env = jinja2.Environment(loader=jinja2.DictLoader({'dict': template}))
     env.filters.update(
         {'arg': lambda x: DeField(x, datavar="fake_dict", name="fake_key")}
