@@ -35,7 +35,11 @@ class MsgPackDeserializer(Deserializer):
 
 
 def to_msgpack(
-    obj: Any, se: Type[Serializer] = MsgPackSerializer, named: bool = True, ext_dict: Dict[Type, int] = None, **opts
+    obj: Any,
+    se: Type[Serializer] = MsgPackSerializer,
+    named: bool = True,
+    ext_dict: Dict[Type, int] = None,
+    **opts,
 ) -> bytes:
     """
     Serialize the object into MsgPack.
@@ -57,7 +61,11 @@ def to_msgpack(
             raise SerdeError(f"Could not find type code for {obj_type.__name__} in ext_dict")
 
     to_func = to_dict if named else to_tuple
-    return se.serialize(to_func(obj, reuse_instances=False, convert_sets=True), ext_type_code=ext_type_code, **opts)
+    return se.serialize(
+        to_func(obj, reuse_instances=False, convert_sets=True),
+        ext_type_code=ext_type_code,
+        **opts,
+    )
 
 
 def from_msgpack(
