@@ -7,12 +7,14 @@ import abc
 import copy
 import dataclasses
 import functools
+import typing
 from dataclasses import dataclass, is_dataclass
 from typing import Any, Callable, Dict, Iterator, List, Optional, Type, TypeVar
 
 import jinja2
 
 from .compat import (
+    Literal,
     SerdeError,
     SerdeSkip,
     get_origin,
@@ -21,15 +23,20 @@ from .compat import (
     is_bare_opt,
     is_bare_set,
     is_bare_tuple,
+    is_datetime,
+    is_datetime_instance,
     is_dict,
     is_enum,
     is_generic,
     is_list,
     is_literal,
+    is_new_type_primitive,
     is_none,
     is_opt,
     is_primitive,
     is_set,
+    is_str_serializable,
+    is_str_serializable_instance,
     is_tuple,
     is_union,
     iter_types,
@@ -41,19 +48,22 @@ from .core import (
     SERDE_SCOPE,
     TO_DICT,
     TO_ITER,
+    TYPE_CHECK,
     UNION_SE_PREFIX,
-    DateTimeTypes,
     DefaultTagging,
     Field,
+    NoCheck,
     SerdeScope,
-    StrSerializableTypes,
     Tagging,
+    TypeCheck,
     add_func,
+    coerce,
     conv,
     fields,
     is_instance,
     logger,
     raise_unsupported_type,
+    render_type_check,
     union_func_name,
 )
 from .numpy import (
@@ -129,6 +139,7 @@ def serialize(
     convert_sets_default: bool = False,
     serializer: Optional[SerializeFunc] = None,
     tagging: Tagging = DefaultTagging,
+    type_check: TypeCheck = NoCheck,
     **kwargs,
 ):
     """
@@ -200,7 +211,7 @@ def serialize(
         # Create a scope storage used by serde.
         # Each class should get own scope. Child classes can not share scope with parent class.
         # That's why we need the "scope.cls is not cls" check.
-        scope: SerdeScope = getattr(cls, SERDE_SCOPE, None)
+        scope: Optional[SerdeScope] = getattr(cls, SERDE_SCOPE, None)
         if scope is None or scope.cls is not cls:
             scope = SerdeScope(
                 cls,
@@ -220,12 +231,17 @@ def serialize(
         g["typename"] = typename  # used in union functions
         g["is_instance"] = is_instance  # used in union functions
         g["to_obj"] = to_obj
+        g["typing"] = typing
+        g["Literal"] = Literal
+        g["TypeCheck"] = TypeCheck
+        g["NoCheck"] = NoCheck
+        g["coerce"] = coerce
         if serialize:
             g["serde_custom_class_serializer"] = functools.partial(serde_custom_class_serializer, custom=serializer)
 
         # Collect types used in the generated code.
         for typ in iter_types(cls):
-            if typ is cls or (is_primitive(typ) and not is_enum(typ)):
+            if typ is cls or (is_primitive(typ) and not is_enum(typ) and not is_new_type_primitive(typ)):
                 continue
             g[typename(typ)] = typ
 
@@ -242,8 +258,9 @@ def serialize(
             if f.serializer:
                 g[f.serializer.name] = f.serializer
 
-        add_func(scope, TO_ITER, render_to_tuple(cls, serializer), g)
-        add_func(scope, TO_DICT, render_to_dict(cls, rename_all, serializer), g)
+        add_func(scope, TO_ITER, render_to_tuple(cls, serializer, type_check), g)
+        add_func(scope, TO_DICT, render_to_dict(cls, rename_all, serializer, type_check), g)
+        add_func(scope, TYPE_CHECK, render_type_check(cls), g)
 
         logger.debug(f"{typename(cls)}: {SERDE_SCOPE} {scope}")
 
@@ -274,7 +291,7 @@ def is_serializable(instance_or_class: Any) -> bool:
     return hasattr(instance_or_class, SERDE_SCOPE)
 
 
-def to_obj(o, named: bool, reuse_instances: bool, convert_sets: bool):
+def to_obj(o, named: bool, reuse_instances: bool, convert_sets: bool, c: Type = None):
     try:
         thisfunc = functools.partial(
             to_obj,
@@ -286,10 +303,8 @@ def to_obj(o, named: bool, reuse_instances: bool, convert_sets: bool):
             return None
         if is_serializable(o):
             serde_scope: SerdeScope = getattr(o, SERDE_SCOPE)
-            if named:
-                return serde_scope.funcs[TO_DICT](o, reuse_instances=reuse_instances, convert_sets=convert_sets)
-            else:
-                return serde_scope.funcs[TO_ITER](o, reuse_instances=reuse_instances, convert_sets=convert_sets)
+            func_name = TO_DICT if named else TO_ITER
+            return serde_scope.funcs[func_name](o, reuse_instances=reuse_instances, convert_sets=convert_sets)
         elif is_dataclass(o):
             if named:
                 return dataclasses.asdict(o)
@@ -303,9 +318,9 @@ def to_obj(o, named: bool, reuse_instances: bool, convert_sets: bool):
             return [thisfunc(e) for e in o]
         elif isinstance(o, dict):
             return {k: thisfunc(v) for k, v in o.items()}
-        elif isinstance(o, StrSerializableTypes):
+        elif is_str_serializable_instance(o):
             return str(o)
-        elif isinstance(o, DateTimeTypes):
+        elif is_datetime_instance(o):
             return o.isoformat()
 
         return o
@@ -410,7 +425,7 @@ def sefields(cls: Type) -> Iterator[SeField]:
         yield f
 
 
-def render_to_tuple(cls: Type, custom: Optional[SerializeFunc] = None) -> str:
+def render_to_tuple(cls: Type, custom: Optional[SerializeFunc] = None, type_check: TypeCheck = NoCheck) -> str:
     template = """
 def {{func}}(obj, reuse_instances = {{serde_scope.reuse_instances_default}},
              convert_sets = {{serde_scope.convert_sets_default}}):
@@ -422,6 +437,10 @@ def {{func}}(obj, reuse_instances = {{serde_scope.reuse_instances_default}},
   if not is_dataclass(obj):
     return copy.deepcopy(obj)
 
+  {% if type_check.is_strict() %}
+  obj.__serde__.funcs['typecheck'](obj)
+  {% endif %}
+
   return (
   {% for f in fields -%}
   {% if not f.skip|default(False) %}
@@ -431,13 +450,17 @@ def {{func}}(obj, reuse_instances = {{serde_scope.reuse_instances_default}},
   )
     """
 
-    renderer = Renderer(TO_ITER, custom)
+    renderer = Renderer(TO_ITER, custom, suppress_coerce=(not type_check.is_coerce()))
     env = jinja2.Environment(loader=jinja2.DictLoader({"iter": template}))
     env.filters.update({"rvalue": renderer.render})
-    return env.get_template("iter").render(func=TO_ITER, serde_scope=getattr(cls, SERDE_SCOPE), fields=sefields(cls))
+    return env.get_template("iter").render(
+        func=TO_ITER, serde_scope=getattr(cls, SERDE_SCOPE), fields=sefields(cls), type_check=type_check
+    )
 
 
-def render_to_dict(cls: Type, case: Optional[str] = None, custom: Optional[SerializeFunc] = None) -> str:
+def render_to_dict(
+    cls: Type, case: Optional[str] = None, custom: Optional[SerializeFunc] = None, type_check: TypeCheck = NoCheck
+) -> str:
     template = """
 def {{func}}(obj, reuse_instances = {{serde_scope.reuse_instances_default}},
              convert_sets = {{serde_scope.convert_sets_default}}):
@@ -449,6 +472,10 @@ def {{func}}(obj, reuse_instances = {{serde_scope.reuse_instances_default}},
   if not is_dataclass(obj):
     return copy.deepcopy(obj)
 
+  {% if type_check.is_strict() %}
+  obj.__serde__.funcs['typecheck'](obj)
+  {% endif %}
+
   res = {}
   {% for f in fields -%}
   {% if not f.skip -%}
@@ -463,13 +490,15 @@ def {{func}}(obj, reuse_instances = {{serde_scope.reuse_instances_default}},
   {% endfor -%}
   return res
     """
-    renderer = Renderer(TO_DICT, custom)
+    renderer = Renderer(TO_DICT, custom, suppress_coerce=(not type_check.is_coerce()))
     lrenderer = LRenderer(case)
     env = jinja2.Environment(loader=jinja2.DictLoader({"dict": template}))
     env.filters.update({"rvalue": renderer.render})
     env.filters.update({"lvalue": lrenderer.render})
     env.filters.update({"case": functools.partial(conv, case=case)})
-    return env.get_template("dict").render(func=TO_DICT, serde_scope=getattr(cls, SERDE_SCOPE), fields=sefields(cls))
+    return env.get_template("dict").render(
+        func=TO_DICT, serde_scope=getattr(cls, SERDE_SCOPE), fields=sefields(cls), type_check=type_check
+    )
 
 
 def render_union_func(cls: Type, union_args: List[Type], tagging: Tagging = DefaultTagging) -> str:
@@ -503,7 +532,7 @@ def {{func}}(obj, reuse_instances, convert_sets):
     """
     union_name = f"Union[{', '.join([typename(a) for a in union_args])}]"
 
-    renderer = Renderer(TO_DICT)
+    renderer = Renderer(TO_DICT, suppress_coerce=True)
     env = jinja2.Environment(loader=jinja2.DictLoader({"dict": template}))
     env.filters.update({"arg": lambda x: SeField(x, "obj")})
     env.filters.update({"rvalue": renderer.render})
@@ -553,6 +582,8 @@ class Renderer:
 
     func: str
     custom: Optional[SerializeFunc] = None  # Custom class level serializer.
+    suppress_coerce: bool = False
+    """ Suppress type coercing because generated union serializer has its own type checking """
 
     def render(self, arg: SeField) -> str:
         """
@@ -560,10 +591,10 @@ class Renderer:
 
         >>> from typing import Tuple
         >>> Renderer(TO_ITER).render(SeField(int, 'i'))
-        'i'
+        'coerce(int, i)'
 
         >>> Renderer(TO_ITER).render(SeField(List[int], 'l'))
-        '[v for v in l]'
+        '[coerce(int, v) for v in l]'
 
         >>> @serialize
         ... @dataclass(unsafe_hash=True)
@@ -576,7 +607,7 @@ class Renderer:
         "[v.__serde__.funcs['to_iter'](v, reuse_instances=reuse_instances, convert_sets=convert_sets) for v in foo]"
 
         >>> Renderer(TO_ITER).render(SeField(Dict[str, Foo], 'foo'))
-        "{k: v.__serde__.funcs['to_iter'](v, reuse_instances=reuse_instances, \
+        "{coerce(str, k): v.__serde__.funcs['to_iter'](v, reuse_instances=reuse_instances, \
 convert_sets=convert_sets) for k, v in foo.items()}"
 
         >>> Renderer(TO_ITER).render(SeField(Dict[Foo, Foo], 'foo'))
@@ -585,8 +616,8 @@ convert_sets=convert_sets): v.__serde__.funcs['to_iter'](v, reuse_instances=reus
 convert_sets=convert_sets) for k, v in foo.items()}"
 
         >>> Renderer(TO_ITER).render(SeField(Tuple[str, Foo, int], 'foo'))
-        "(foo[0], foo[1].__serde__.funcs['to_iter'](foo[1], reuse_instances=reuse_instances, \
-convert_sets=convert_sets), foo[2],)"
+        "(coerce(str, foo[0]), foo[1].__serde__.funcs['to_iter'](foo[1], reuse_instances=reuse_instances, \
+convert_sets=convert_sets), coerce(int, foo[2]),)"
         """
         if arg.serializer and arg.serializer.inner is not default_serializer:
             res = self.custom_field_serializer(arg)
@@ -614,9 +645,9 @@ convert_sets=convert_sets), foo[2],)"
             res = self.primitive(arg)
         elif is_union(arg.type):
             res = self.union_func(arg)
-        elif arg.type in StrSerializableTypes:
+        elif is_str_serializable(arg.type):
             res = f"{arg.varname} if reuse_instances else {self.string(arg)}"
-        elif arg.type in DateTimeTypes:
+        elif is_datetime(arg.type):
             res = f"{arg.varname} if reuse_instances else {arg.varname}.isoformat()"
         elif is_none(arg.type):
             res = "None"
@@ -729,7 +760,12 @@ convert_sets=convert_sets), foo[2],)"
         """
         Render rvalue for primitives.
         """
-        return f"{arg.varname}"
+        typ = typename(arg.type)
+        var = arg.varname
+        if self.suppress_coerce:
+            return var
+        else:
+            return f'coerce({typ}, {var})'
 
     def string(self, arg: SeField) -> str:
         return f"str({arg.varname})"
